@@ -3,7 +3,8 @@
 # Eventually:
 #    the other documentation will refer to this file
 #    there will be another file with the fixed function
-
+use strict;
+use warnings;
 
 
 
@@ -29,13 +30,11 @@ sub binary_search {
   }
 }
 
-$Devel::DumpTrace::LEVEL = 0;
-
 # at first glance it looks correct but it fails on some test cases
 
-@w = sort qw(the quick brown fox jumps over the lazy dog);
-for $w (@w) {
-  $ww = binary_search($w, @w);
+my @w = sort qw(the quick brown fox jumps over the lazy dog);
+for my $w (@w) {
+  my $ww = binary_search($w, @w);
   if ($w[$ww] eq $w) { 
     print "$ww $w OK\n";
   } else { 
@@ -48,15 +47,10 @@ for $w (@w) {
 # Let's debug it with Devel::DumpTrace
 
 BEGIN { $Devel::DumpTrace::TRACE = 0 };
+if ($INC{"Devel/DumpTrace.pm"}) {
+    local $Devel::DumpTrace::TRACE = 'normal';
+    my $ww = binary_search($w[4], @w);
+    print "\$ww is $ww, should be 4.\n";
+}
 
-
-$Devel::DumpTrace::TRACE = 'verbose';
-  
-$ww = binary_search($w[4], @w);
-print "\$ww is $ww, should be 4.\n";
-
-$Devel::DumpTrade::TRACE = 0;
-
-# Examine the output. Can you spot the error? Can you fix it?
-
-
+# Examine the output. Can you spot the logic error? Can you fix it?
